/*
 * iwconnect/iwconnect.c
 * $Rev: 1739 $
 *
 * Copyright (C) 2008-2010, Reto Gantenbein <reto.gantenbein@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file iwconnect.c
 * @brief iwconnect - Traffic redirection deamon for virtual wireless devices
 *
 * Usage: iwconnect WLANMODEL [-i INTERFACE][-p PORT][-6][-n][-d]
 *
 * Tunnels traffic from a virtual wireless interface to a
 * wireless emulation model. First you have to create a virtual
 * interface with `vifctl -c INTERFACE`. The virtual interfaces are
 * then detected rm automatically.
 *
 * Parameters:
 *
 * WLANMODEL    IP Address/Hostname and port number of wireless model
 * LOCALPORT    Local port to listen for model replies
 * -6           Use IPv6 protocol
 * -d           Enable debug output
 * -p PORT      Destination port (default: 2424)
 * -i INTERFACE	Tunnel interface (default: eth0)
 * -n           Enable direct node-to-node communication
 * -h           Show this help message
 */

/* INCLUDES *******************************************************************/

#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>                 // va_list, va_start(), va_end(), vfprintf(), ...
#include <stdio.h>                  // printf(), fprintf()
#include <stdlib.h>                 // exit()
#include <string.h>                 // memset(), strcpy(), ...
#include <syslog.h>                 // openlog(), vsyslog(), LOG_PID, ...
#include <unistd.h>                 // getopt(), optarg, getuid()
#include <sys/resource.h>           // getpriority()
#include <sys/types.h>
#include "netdevice.h"              // init_udp_socket()
#include "message.h"
#include "iwconnect.h"
#include "iwc_print.h"


/* EXTERNAL FUNCTIONS *********************************************************/

/* iwc_init.c */
extern int iwc_init_host();
extern int iwc_init_model(const char *, int);
extern int iwc_init_net();
extern int iwc_init_check_pidfile(const char *);
extern int iwc_init_create_pidfile(const char *);
extern int iwc_init_remove_pidfile(const char *);
/* iwc_model.c */
extern int iwc_model_register();
extern int iwc_model_deregister();
extern void iwc_model_update(int, const char*);
/* iwc_notify.c */
extern void iwc_notify_poll(void *);


/* PRIVATE VARIABLES **********************************************************/

char *exec_name = NULL;                     /**< Program name */
const char iwc_version[] = "1.0\0";          /**< Program version */
const char iwc_pidfile[] = "/var/run/iwconnect.pid";   /**< PID file */

/* Host parameters */
iwc_client_info client_info;                 /**< Wireless client information */
iwc_model_info model_info;                   /**< Wireless model information */

/** Number of virtual interfaces */
int num_vif;

/** Names of virtual interfaces */
char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1];

/* Socket file descriptors */
int tap_sockets[MAX_NUM_VIF];               /**< TAP socket file descriptors */
int net_socket;                             /**< Network socket file descriptor */

/** Indicate direct node to node mode */
int direct;

/** Indicate debug level (0|1) */
int iwc_dbg;

/** Are we registered at the WlanModel */
int is_registered;

/** IP protocol version to use (AF_INET for IPv4, AF_INET6 for IPv6) */
int ip_family;


/* FUNCTIONS ******************************************************************/

/** Print usage details */
void
iwc_usage()
{
    printf("\nUsage: %s WLANMODEL [-p PORT][-i INTERFACE][-6][-n][-d][-h]\n", exec_name);
    printf("Try `%s -h' for more details\n", exec_name);

    exit(EXIT_FAILURE);
}

/** Print help */
void
iwc_help()
{
    printf("Usage: %s WLANMODEL [-p PORT][-i INTERFACE][-6][-n][-d][-h]\n", exec_name);
    printf("\nIwconnect tunnels traffic from a virtual wireless interface\n");
    printf("to a wireless emulation model. First you have to create a virtual\n");
    printf("interface with `vifctl -c INTERFACE`. The virtual interfaces are \n");
    printf("then detected automatically by iwconnect.\n");
    printf("\nParameters:\n");
    printf("WLANMODEL\tIP Address/Hostname and port number of wireless model\n");
    printf("LOCALPORT\tLocal port to listen for model replies\n");
    printf("-p PORT\t\tDestination port (default: 2424)\n");
    printf("-i INTERFACE\tInfrastructure network interface (default: %s)\n", client_info.host_if);
    printf("-6\t\tUse IPv6 protocol\n");
    printf("-d\t\tEnable debug output\n");
    printf("-n\t\tEnable node-to-node communication (only for testing!)\n");
    printf("-h\t\tShow this help message\n");
    exit(EXIT_SUCCESS);
}

/** Stub for print_debug signature */
int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Print debug message with variable argument list */
int
print_debug(const char *format, ...)
{
    va_list arg;

    va_start(arg, format);
    iwc_debug_args(format, arg);
    va_end(arg);

    return 0;
}

/** Stub for print_debug_args signature */
void
ignore_debug_args(const char *format, va_list arg)
{
}

/** Print debug message with given argument list */
void
print_debug_args(const char *format, va_list arg)
{
    FILE *stdout = fopen("/dev/stdout", "a");
    fprintf(stdout, "DEBUG: ");
    vfprintf(stdout, format, arg);
    fclose(stdout);
}

/** Print error message */
int
iwc_error(const char *format, ...)
{
    int chars_printed = 0;
    va_list arg;

    FILE *stderr = fopen("/dev/stderr", "a");
    if (stderr == NULL)
        return -1;
    chars_printed = fprintf(stderr, "Error: ");
    va_start(arg, format);
    chars_printed += vfprintf(stderr, format, arg);
    va_end(arg);
    fclose(stderr);

    // also send to syslog
    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    va_end(arg);
    closelog();

    return chars_printed;
}

/** Send message to syslog and print to stdout */
void
iwc_log(const char *format, ...)
{
    va_list arg;

    // send to syslog
    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    va_end(arg);
    closelog();

    // also print to stdout
    FILE *stdout = fopen("/dev/stdout", "a");
    if (stdout == NULL)
        return;
    va_start(arg, format);
    vfprintf(stdout, format, arg);
    va_end(arg);
    fclose(stdout);
}

/** Shutdown iwconnect and deregister at WlanModel */
void
iwc_shutdown(int signal)
{
    int i;
    int rtrn = 0;

    iwc_log("iwconnect: Shutdown\n");

    // deregister model
    if (is_registered == 1)
    {
        if (iwc_model_deregister() < 0)
            signal = rtrn;          // TODO: WTF?!
        else
            is_registered = 0;
    }

    // close network sockets
    close(net_socket);
    for (i = 0; i < num_vif; i++)
        close(tap_sockets[i]);

    // remove pid file
    iwc_init_remove_pidfile(iwc_pidfile);

    if (((signal == SIGINT) ||
        (signal == SIGABRT) ||
        (signal == SIGTERM)) &&
        (rtrn == 0))
    {
        exit(EXIT_SUCCESS);
    } else {
        exit(signal);
    }
}


/** Listen to TUN and network interfaces for "wireless" packets */
void
iwc_daemon()
{
    fd_set fds;
    int i, recv_bytes, sent_bytes;
    register int ret;
    char buffer[DATA_MSG_SIZE];                     // biggest possible message
    struct data_msg *message = (struct data_msg *) buffer;
    raw_packet *ether_frame = (raw_packet *) message->wlan_pkt;

    iwc_debug("iwc_daemon(): Running main loop with %i vif(s)\n", num_vif);
    iwc_debug("\n");
    iwc_init_create_pidfile(iwc_pidfile);   // create pid file with forked pid

    // find biggest fd, for select()
    int max_fd = 0;
    for (i = 0; i < num_vif; i++)
        max_fd = max(max_fd, tap_sockets[i]);
    max_fd = max(max_fd, net_socket) + 1;

    for (;;)
    {
        FD_ZERO(&fds);
        for (i = 0; i < num_vif; i++)
            FD_SET(tap_sockets[i], &fds);
        FD_SET(net_socket, &fds);

        // check for fd with pending input data
        if ((ret = select(max_fd, &fds, NULL, NULL, NULL)) < 0)
        {
            iwc_error("iwc_daemon(): select() (%s)\n",
                    strerror(errno));
            return;
        } else {
            // check tap device for incoming traffic
            for (i = 0; i < num_vif; i++)
            {
                if (FD_ISSET(tap_sockets[i], &fds))
                {
                    // read packet into wlan_pkt argument of wlan_msg
                    recv_bytes = read(tap_sockets[i], ether_frame,
                                    MAX_FRAME_SIZE);
                    iwc_debug(">>tap_fd<< %i Bytes read\n", recv_bytes);
                    if (recv_bytes > 0)
                    {
                        message->header.type = DATA_MSG;
                        memmove(&message->header.id, client_info.host_id,
                                sizeof(client_info.host_id));
                        message->pkt_length = recv_bytes;
                        message->vif_index = i;
                        debug_msg((char *) message);

                        // send data to model
                        if ((sent_bytes = sendto(net_socket, message,
                                DATA_PREAMBLE_SIZE + recv_bytes, 0,
                                model_info.host_info.addr_info.ai_addr,
                                model_info.host_info.addr_info.ai_addrlen)) < 0)
                        {
                            iwc_error("iwc_daemon(): net:sendto (%s)\n",
                                    strerror(errno));
                        }
                        iwc_debug(">>net_socket<< %i Bytes sent\n", sent_bytes);
                        iwc_debug("\n");
                    } else
                        iwc_error("iwc_daemon(): tap:read (%s)\n", strerror(errno));
                }
            }
            // check network device for incoming traffic
            if (FD_ISSET(net_socket, &fds))
            {
                recv_bytes = recvfrom(net_socket, buffer, DATA_MSG_SIZE, 0,
                                    NULL, NULL);
                iwc_debug(">>net_socket<< %i Bytes read\n", recv_bytes);
                if (recv_bytes > 0)
                {
                    if (is_data_msg((struct msg_header *) buffer))
                    {
                        debug_msg(buffer);

                        // send to TAP interface
                        iwc_debug("iwc_daemon(): sendto fd = %i, pkt_length = %i\n",
                                tap_sockets[message->vif_index],
                                message->pkt_length);
                        if ((sent_bytes = write(
                                tap_sockets[message->vif_index],
                                ether_frame, message->pkt_length)) < 1)
                        {                       // handle write() error
                            iwc_error("iwc_daemon(): write() (%s)\n",
                                    strerror(errno));
                        } else {
                            iwc_debug(">>tap_fd<< %i Bytes sent\n", sent_bytes);
                            iwc_debug("\n");
                        }
                    } else {
                        if (is_ack_msg((struct msg_header *) buffer))
                        {
                            iwc_debug("iwc_daemon(): Discard straying Ack message\n");
                            iwc_debug("\n");
                        } else {
                            iwc_debug("iwc_daemon(): Unknown message received\n");
                            iwc_debug("\n");
                        }
                    }
                } else {                         // handle recvfrom() errors
                    if (errno == ECONNREFUSED)
                        iwc_log("No connection to WlanModel (%s)\n", strerror(errno));
                    else
                        iwc_error("iwc_daemon(): recvfrom() (%s)\n", strerror(errno));
                    return;
                }
            }
        }
    }
}


/** Parse command line options and checks it for validity */
int
iwc_pars_opts(int argc, char *argv[])
{
    int opt = -1;
    int debug = 0;
    int model_port = -1;
    char *local_port = NULL;

    // IMPORTANT: Don't use iwc_log or iwc_debug before here!
    iwc_debug = ignore_debug;
    iwc_debug_args = ignore_debug_args;

    if (argc < 2)
    {
        iwc_error("Missing arguments\n");
        return -1;
    }

    // create a channel to the NET kernel
    if (init_udp_socket() < 0)
        return (EXIT_FAILURE);

    while ((opt = getopt(argc, argv, "6dnp:hi:")) != -1)
    {
        switch (opt)
        {
            case 'h':
                iwc_help();
                break;
            case 'i':
                if (get_ip_by_dev(NULL, optarg)) // check if configured interface
                    return -1;
                else
                {   // we cannot trust the input, correctly create string
                    strncpy(client_info.host_if, optarg, IFNAMSIZ);
                    client_info.host_if[IFNAMSIZ] = '\0';
                }
                break;
            case 'd':
                debug = 1;
                break;
            case 'n':
                direct = 1;
                break;
            case 'p':
                model_port = atoi(optarg);  // first store in normal int to check range
                if ((model_port > 65536) || (model_port < 1))
                {
                    iwc_error("Invalid local port number\n");
                    return -1;
                }
                break;
            case '6':
                ip_family = PF_INET6;
                break;
        }
    }

    // configure debug/error streams
    if (debug == 1)
    {
        iwc_dbg = 1;
        iwc_debug = print_debug;
        iwc_debug_args = print_debug_args;
        vif_set_debug_function(iwc_debug);
    }
    vif_set_error_function(iwc_error);

    if (debug == 1)
        iwc_debug("Debugging enabled!\n");
    if (ip_family == AF_INET6)
        iwc_log("Using IPv6 protocol\n");

    // set local port
    local_port = getenv("IWCPORT");
    if (local_port == NULL)
        local_port = "2424";
    else
        if (strlen(local_port) == 0)
            local_port = "2424";
    memmove(client_info.host_port, local_port, 5);

    // read model address (last argument, because we used getopt)
    if (iwc_init_model(argv[argc - 1], model_port) == -1)
    {
        iwc_error("Model initialization failed!\n");
        return -1;
    }
    iwc_debug("Using local port %s\n", local_port);
    iwc_debug("Using local interface %s\n", client_info.host_if);

    return 0;
}

/** MAIN */
int
main(int argc, char *argv[])
{
    pthread_t tid;
    pid_t pid = 0;
    int is_running = -1;
    int user_nice = 0;

    // extract executable name
    if (strrchr(argv[0], '/') == NULL)
    {
        exec_name = argv[0];
    } else {
        exec_name = strrchr(argv[0], '/') + 1;
    }

    // initialize default values
    num_vif = 0;
    ip_family = AF_INET;                    // AF_INET (ipv4), AF_INET6 (ipv6)
    direct = 0;
    iwc_dbg = 0;
    is_registered = 0;
    memset(&model_info, 0, sizeof(iwc_model_info));
    memset(&client_info, 0, sizeof(iwc_client_info));
    strcpy(client_info.host_if, "eth0\0");            // set default interface

    // check parameters
    if (iwc_pars_opts(argc, argv) < 0)
        iwc_usage();

    // check pid file
    is_running = iwc_init_check_pidfile(iwc_pidfile);
    if (is_running != 0)
        return EXIT_FAILURE;

    iwc_log("Starting iwconnect %s\n", &iwc_version);

    // initialize host values
    if (iwc_init_host() < 0)
    {
        iwc_error("Host initialization failed!\n");
        return EXIT_FAILURE;
    }

    // registering signal handler
    signal(SIGINT, iwc_shutdown);
    signal(SIGABRT, iwc_shutdown);
    signal(SIGTERM, iwc_shutdown);

    // initialize network sockets
    if (iwc_init_net() < 0)
    {
        iwc_error("Network initialization failed!\n");
        return EXIT_FAILURE;
    }

    iwc_debug("Successfully finished host initialization!\n");
    iwc_debug("\n");

    // register node at WlanModel
    if (direct != 1)
    {
        if ((is_registered = iwc_model_register()) != 1)
            return EXIT_FAILURE;
    } else {
        iwc_log("Not registering host!\n");
    }

    // if not debug, fork to execute in background
    if (iwc_dbg == 0)
    {
        pid = fork();
        if (pid < 0)
        {
            iwc_error("Couldn't fork process to background\n");
            iwc_shutdown(EXIT_FAILURE);
        }
        if (pid > 0)
            return EXIT_SUCCESS;    // exit parent process and release console
    }

    // create thread for updating model at vif configuration changes
    if (pthread_create(&tid, NULL, iwc_notify_poll,
        (void *) &iwc_model_update) < 0)
    {
        iwc_error("Couldn't create new thread (%s)\n", strerror(errno));
        iwc_shutdown(EXIT_FAILURE);
    }

    // set high priority nice value
    if (setpriority(PRIO_PROCESS, 0, -20) == 0)
        iwc_log("Set nice priority to %i\n", getpriority(PRIO_PROCESS, 0));
    else
        iwc_log("Failed to set new priority. Running with nice priority %i\n", getpriority(PRIO_PROCESS, 0));

    // run packet redirecting daemon
    iwc_daemon();

    exit(EXIT_FAILURE);
}
